#include "version.h"

#include <malloc.h>
#include <Windows.h>

#define _countof(_Array) (sizeof(_Array) / sizeof(_Array[0]))

// Generic text processing /////////////////////////////////////////////////////

static BOOL findNextToken(PCWSTR *ppsz, size_t *pcChars)
{
    PCWSTR psz = *ppsz;
    size_t iChar;
    size_t i1stChar;
    BOOL bQuotOpen;

    iChar = *pcChars;
    for (;;)
    {
        WCHAR ch = psz[iChar];

        if (ch == '\0') return FALSE;

        if (ch != ' ')
        {
            i1stChar = iChar;
            bQuotOpen = ch == '\"';
            break;
        }
        ++iChar;
    }
    for (;;)
    {
        WCHAR ch = psz[++iChar];

        if (ch == '\0') break;

        if (ch == '\"') bQuotOpen = !bQuotOpen;
        if (!bQuotOpen && (ch == ' ' || ch == '/')) break;
    }
    *ppsz = psz + i1stChar;
    *pcChars = iChar - i1stChar;
    return TRUE;
}

static PWSTR
formatUnitsAndCents(
    unsigned long long ullUnits,
    unsigned int uCents,
    PWSTR psz)
{
    wsprintfW(psz, L"%I64u.%02u", ullUnits, uCents);
    return psz;
}

static size_t getUnquotedLength(const WCHAR *ps, size_t cChars)
{
    size_t iChar;
    size_t cCharsUnquoted;

    cCharsUnquoted = 0;
    for (iChar = 0; iChar < cChars; ++iChar)
    {
        WCHAR ch = ps[iChar];

        if (ch != '\"') ++cCharsUnquoted;
    }
    return cCharsUnquoted;
}

static BOOL isOptionPrefix(WCHAR ch)
{
    return ch == '/' || ch == '-';
}

static void
pickOption(
    const WCHAR *psArg,
    size_t cCharsArg,
    const WCHAR **ppsOpt,
    size_t *pcCharsOpt)
{
    *ppsOpt = psArg + 1;
    *pcCharsOpt = cCharsArg - 1;
}

static void unquote(PWSTR pszDest, const WCHAR *psSrc, size_t cCharsSrc)
{
    size_t iCharDest = 0;
    {
        size_t iCharSrc;

        for (iCharSrc = 0; iCharSrc < cCharsSrc; ++iCharSrc)
        {
            WCHAR ch = psSrc[iCharSrc];

            if (ch != '\"') pszDest[iCharDest++] = ch;
        }
    }
    pszDest[iCharDest] = '\0';
}

static wchar_t *wmemcpy(wchar_t *dest, const wchar_t *src, size_t count)
{
    size_t index;

    for (index = 0; index < count; ++index) dest[index] = src[index];
    return dest;
}

static wchar_t *wmemset(wchar_t *dest, wchar_t c, size_t count)
{
    size_t index;

    for (index = 0; index < count; ++index) dest[index] = c;
    return dest;
}

// Application specific ////////////////////////////////////////////////////////

#define EXIT_CODE_OK        0
#define EXIT_CODE_IDENTICAL 0
#define EXIT_CODE_DIFFERENT 1
#define EXIT_CODE_ERROR     2

#define OPT_UNKNOWN             0
#define OPT_QUIET               0x01
#define OPT_TIMING              0x02
#define OPT_HEED_REPARSE_POINTS 0x04

#define HELP_TEXT \
L"\n\
BINEQ " _CRT_WIDE(STR_VERSION) L" - Copyright " _CRT_WIDE(STR_YEAR_AUTHOR) \
L" - " _CRT_WIDE(STR_COMPANY) L"\n\
\n\
Tests if the contents of two or more files are identical.\n\
\n\
Usage: BINEQ [/Q] [/T] [/R] file1 file2 [file3 [...]]\n\
\n\
  /Q                           Do not print comparison results.\n\
  /T                           Print timing information.\n\
  /R                           Also compare reparse points.\n\
  file1, file2, file3, ...     Files to be compared.\n\
\n\
ERRORLEVEL is set to 0 if the data in all files are identical, or to 1 if they\
\n\
are not. If an error occurs, ERRORLEVEL is set to 2.\n\
"

#define MAX_PROGRESS_DIGITS         2

#define WSTR_MAX_PROGRESS_DIGITS    \
_CRT_WIDE(_CRT_STRINGIZE(MAX_PROGRESS_DIGITS))

#define ZERO_PROGRESS_WCHAR_ARRAY   L' ', L'0'

typedef enum
{
    ARG_TYPE_OPTION,
    ARG_TYPE_FILE_NAME,
}
ARG_TYPE;

typedef enum
{
    FILE_DIFF_NONE,
    FILE_DIFF_SIZE,
    FILE_DIFF_CONTENTS,
    FILE_DIFF_REPARSE_POINTS,
}
FILE_DIFF;

struct FileInfo;

typedef struct FileInfo FILE_INFO;

#pragma warning(push)
#pragma warning(disable: 4200 4201)

struct FileInfo
{
    FILE_INFO *pNext;
    union
    {
        HANDLE hFile;
        HANDLE hFileMapping;
    };
    WCHAR szFileName[];
};

#pragma warning(pop)

static BOOL isArgumentOptionHelp(const WCHAR *psArg, size_t cChars)
{
    return cChars == 2 && isOptionPrefix(psArg[0]) && psArg[1] == '?';
}

static HANDLE getStdHandle(DWORD dwStdHandle)
{
    HANDLE hStd;

    hStd = GetStdHandle(dwStdHandle);
    if (hStd == INVALID_HANDLE_VALUE) hStd = NULL;
    return hStd;
}

static BOOL write(PCWSTR psz, HANDLE hFile)
{
    BOOL bRes;
    {
        // The type casts of cch and size from size_t to DWORD are safe since we
        // are only using strings with known maximum lengths.
        size_t size;
        PSTR psOem;
        {
            size_t cch = wcslen(psz);

            size = cch * sizeof(CHAR);
            psOem = alloca(size);
            CharToOemBuffW(psz, psOem, (DWORD)cch);
        }
        {
            DWORD dwNumberOfBytesWritten;

            bRes =
                WriteFile(
                hFile,
                psOem,
                (DWORD)size,
                &dwNumberOfBytesWritten,
                NULL
                );
        }
    }
    return bRes;
}

static __declspec(noreturn) void writeErr(PCWSTR psz, ...)
{
    {
        HANDLE hStdError;

        hStdError = getStdHandle(STD_ERROR_HANDLE);
        if (hStdError != NULL)
        {
            va_list args;
            PCWSTR pszW;

            va_start(args, psz);
            for (pszW = psz; pszW != NULL; pszW = va_arg(args, PCWSTR))
            {
                BOOL bRes = write(pszW, hStdError);

                if (!bRes) break;
            }
        }
    }
    ExitProcess(EXIT_CODE_ERROR);
}

static void writeOut(PCWSTR psz, ...)
{
    static HANDLE hStdOutput_ = INVALID_HANDLE_VALUE;

    if (hStdOutput_ == INVALID_HANDLE_VALUE)
        hStdOutput_ = getStdHandle(STD_OUTPUT_HANDLE);
    if (hStdOutput_ != NULL)
    {
        va_list args;
        PCWSTR pszW;

        va_start(args, psz);
        for (pszW = psz; pszW != NULL; pszW = va_arg(args, PCWSTR))
        {
            BOOL bRes = write(pszW, hStdOutput_);

            if (!bRes)
            {
                hStdOutput_ = NULL;
                break;
            }
        }
    }
}

static ARG_TYPE getArgumentType(const WCHAR *psArg)
{
    WCHAR ch;

    ch = psArg[0];
    return isOptionPrefix(ch) ? ARG_TYPE_OPTION : ARG_TYPE_FILE_NAME;
}

static DWORD getOption(const WCHAR *psOpt, size_t cCharsOpt)
{
    if (cCharsOpt == 1)
    {
        WCHAR ch;
        
        ch = psOpt[0];
        switch (ch)
        {
        case 'Q':
        case 'q':
            return OPT_QUIET;
        case 'T':
        case 't':
            return OPT_TIMING;
        case 'R':
        case 'r':
            return OPT_HEED_REPARSE_POINTS;
        }
    }
    return OPT_UNKNOWN;
}

static void
appendFileInfo(
    const WCHAR *psFileName,
    size_t cCharsFileName,
    FILE_INFO **ppfileInfoLast)
{
    FILE_INFO *pfileInfo;
    {
        SIZE_T uBytes;
        {
            size_t cCharsFileNameUnquoted =
                getUnquotedLength(psFileName, cCharsFileName);

            uBytes =
                (sizeof(FILE_INFO) + sizeof(WCHAR)) +
                cCharsFileNameUnquoted * sizeof(WCHAR);
        }
        pfileInfo = (FILE_INFO *)LocalAlloc(LMEM_FIXED, uBytes);
    }
    if (pfileInfo == NULL) writeErr(L"Not enough memory.\n", NULL);
    unquote(pfileInfo->szFileName, psFileName, cCharsFileName);
    {
        FILE_INFO *pNext;
        {
            FILE_INFO *pfileInfoLast = *ppfileInfoLast;

            if (pfileInfoLast != NULL)
            {
                pNext = pfileInfoLast->pNext;
                pfileInfoLast->pNext = pfileInfo;
            }
            else
                pNext = pfileInfo;
        }
        pfileInfo->pNext = pNext;
    }
    *ppfileInfoLast = pfileInfo;
}

#pragma warning(push)
#pragma warning(disable: 4701 4703)

static void
parseCommandLine(FILE_INFO **ppfileInfo1st, DWORD *pdwOptSet)
{
    FILE_INFO *pfileInfoLast = NULL;
    unsigned int cFiles;

    cFiles = 0;
    *pdwOptSet = 0;
    {
        PCWSTR pszErrorText;
        PWSTR pszBadOpt;

        pszErrorText = NULL;
        {
            PCWSTR psz = GetCommandLineW();
            size_t cChars = 0;

            findNextToken(&psz, &cChars);
            while (findNextToken(&psz, &cChars))
            {
                if (isArgumentOptionHelp(psz, cChars))
                {
                    writeOut(HELP_TEXT, NULL);
                    ExitProcess(EXIT_CODE_OK);
                }
                else if (pszErrorText == NULL)
                {
                    switch (getArgumentType(psz))
                    {
                    case ARG_TYPE_OPTION:
                        {
                            const WCHAR *psOpt;
                            size_t cCharsOpt;
                            DWORD dwOpt;

                            pickOption(psz, cChars, &psOpt, &cCharsOpt);
                            dwOpt = getOption(psOpt, cCharsOpt);
                            if (dwOpt == OPT_UNKNOWN)
                                pszErrorText = L"Invalid option";
                            else if (dwOpt & *pdwOptSet)
                                pszErrorText = L"Duplicated option";
                            else
                            {
                                *pdwOptSet |= dwOpt;
                                break;
                            }
                            pszBadOpt = alloca((cCharsOpt + 1) * sizeof(WCHAR));
                            wmemcpy(pszBadOpt, psOpt, cCharsOpt);
                            pszBadOpt[cCharsOpt] = '\0';
                        }
                        break;
                    case ARG_TYPE_FILE_NAME:
                        appendFileInfo(psz, cChars, &pfileInfoLast);
                        ++cFiles;
                        break;
                    default:
                        __assume(0);
                    }
                }
            }
        }
        if (pszErrorText != NULL)
            writeErr(
            pszErrorText,
            L" \"",
            pszBadOpt,
            L"\".\n",
            NULL
            );
    }
    if (cFiles < 2) writeErr(L"At least two files must be specified.\n", NULL);
    *ppfileInfo1st = pfileInfoLast->pNext;
}

#pragma warning(pop)

static HANDLE openFile(PCWSTR pszFileName, DWORD dwFlagsAndAttributes)
{
    return
        CreateFileW(
        pszFileName,
        GENERIC_READ,
        FILE_SHARE_READ,
        NULL,
        OPEN_EXISTING,
        dwFlagsAndAttributes,
        NULL
        );
}

#ifndef _WIN64

#define OPEN_FILE(                          \
    pszFileName,                            \
    dwFlagsAndAttributes,                   \
    pfnWow64DisableWow64FsRedirection,      \
    pfnWow64RevertWow64FsRedirection)       \
    openFileWin32(                          \
    (pszFileName),                          \
    (dwFlagsAndAttributes),                 \
    (pfnWow64DisableWow64FsRedirection),    \
    (pfnWow64RevertWow64FsRedirection))

typedef BOOL (WINAPI *Wow64DisableWow64FsRedirectionProc)(PVOID *);
typedef BOOL (WINAPI *Wow64RevertWow64FsRedirectionProc)(PVOID);

#pragma warning(push)
#pragma warning(disable: 4701)

static HANDLE
    openFileWin32(
    PCWSTR pszFileName,
    DWORD dwFlagsAndAttributes,
    Wow64DisableWow64FsRedirectionProc pfnWow64DisableWow64FsRedirection,
    Wow64RevertWow64FsRedirectionProc pfnWow64RevertWow64FsRedirection)
{
    PVOID pvOldValue;
    HANDLE hFile;

    if (pfnWow64DisableWow64FsRedirection != NULL)
        pfnWow64DisableWow64FsRedirection(&pvOldValue);
    hFile = openFile(pszFileName, dwFlagsAndAttributes);
    if (pfnWow64RevertWow64FsRedirection != NULL)
        pfnWow64RevertWow64FsRedirection(pvOldValue);
    return hFile;
}

#pragma warning(pop)

#else // #ifndef _WIN64

#define OPEN_FILE(                          \
    pszFileName,                            \
    dwFlagsAndAttributes,                   \
    pfnWow64DisableWow64FsRedirection,      \
    pfnWow64RevertWow64FsRedirection)       \
    openFile((pszFileName), (dwFlagsAndAttributes))

#endif // #ifndef _WIN64

static void openFiles(FILE_INFO *pfileInfo1st, BOOL bHeedReparsePoints)
{
    FILE_INFO *pfileInfo;
    DWORD dwFlagsAndAttributes;

#ifndef _WIN64

    Wow64DisableWow64FsRedirectionProc pfnWow64DisableWow64FsRedirection;
    Wow64RevertWow64FsRedirectionProc pfnWow64RevertWow64FsRedirection;

    {
        HMODULE hModule = GetModuleHandleW(L"KERNEL32.dll");

        pfnWow64DisableWow64FsRedirection =
            (Wow64DisableWow64FsRedirectionProc)
            GetProcAddress(hModule, "Wow64DisableWow64FsRedirection");
        pfnWow64RevertWow64FsRedirection =
            (Wow64RevertWow64FsRedirectionProc)
            GetProcAddress(hModule, "Wow64RevertWow64FsRedirection");
    }

#endif // #ifndef _WIN64

    pfileInfo = pfileInfo1st;
    dwFlagsAndAttributes =
        bHeedReparsePoints ?
        FILE_FLAG_SEQUENTIAL_SCAN | FILE_FLAG_OPEN_REPARSE_POINT :
        FILE_FLAG_SEQUENTIAL_SCAN;
    do
    {
        PCWSTR pszFileName = pfileInfo->szFileName;
        HANDLE hFile =
            OPEN_FILE(
            pszFileName,
            dwFlagsAndAttributes,
            pfnWow64DisableWow64FsRedirection,
            pfnWow64RevertWow64FsRedirection
            );

        if (hFile == INVALID_HANDLE_VALUE)
            writeErr(
            L"The file \"",
            pszFileName,
            L"\" cannot be opened.\n",
            NULL
            );
        pfileInfo->hFile = hFile;
        pfileInfo = pfileInfo->pNext;
    }
    while (pfileInfo != pfileInfo1st);
}

static __declspec(noreturn) void writeErrFileRead(const FILE_INFO *pfileInfo)
{
    PCWSTR pszFileName = pfileInfo->szFileName;

    writeErr(L"The file \"", pszFileName, L"\" cannot be read.\n", NULL);
}

static LONGLONG getFileSize(const FILE_INFO *pfileInfo)
{
    LARGE_INTEGER fileSize;

    if (!GetFileSizeEx(pfileInfo->hFile, &fileSize))
        writeErrFileRead(pfileInfo);
    return fileSize.QuadPart;
}

static int
readReparseData(const FILE_INFO *pfileInfo, void *pvBuffer, DWORD dwBufferSize)
{
    BOOL bRes;
    DWORD dwBytesReturned;
    DWORD dwLastError;

    bRes =
        DeviceIoControl(
        pfileInfo->hFile,
        FSCTL_GET_REPARSE_POINT,
        NULL,
        0,
        pvBuffer,
        dwBufferSize,
        &dwBytesReturned,
        NULL
        );
    if (bRes) return dwBytesReturned;

    dwLastError = GetLastError();
    switch (dwLastError)
    {
    case ERROR_MORE_DATA:
        return -1;
    case ERROR_NOT_A_REPARSE_POINT:
        return 0;
    default:
        writeErrFileRead(pfileInfo);
    }
}

static BOOL compareReparsePoints(const FILE_INFO *pfileInfo1st)
{
    BYTE buffer0[MAXIMUM_REPARSE_DATA_BUFFER_SIZE];
    int nBytesReturned0;
    DWORD dwBufferSize;
    void *pvBuffer;
    const FILE_INFO *pfileInfo;

    nBytesReturned0 = readReparseData(pfileInfo1st, buffer0, sizeof buffer0);
    dwBufferSize = max(nBytesReturned0, REPARSE_GUID_DATA_BUFFER_HEADER_SIZE);
    pvBuffer = alloca(dwBufferSize);
    for (pfileInfo = pfileInfo1st; ; )
    {
        int nBytesReturned;

        pfileInfo = pfileInfo->pNext;
        if (pfileInfo == pfileInfo1st) break;
    
        nBytesReturned = readReparseData(pfileInfo, pvBuffer, dwBufferSize);
        if (
            nBytesReturned != nBytesReturned0 ||
            memcmp(buffer0, pvBuffer, nBytesReturned0) != 0)
            return FALSE;
    }
    return TRUE;
}

static void createFileMappingsAndCloseFiles(FILE_INFO *pfileInfo1st)
{
    FILE_INFO *pfileInfo;

    pfileInfo = pfileInfo1st;
    do
    {
        HANDLE hFile = pfileInfo->hFile;
        HANDLE hFileMapping =
            CreateFileMappingW(
            hFile,
            NULL,
            PAGE_READONLY,
            0,
            0,
            NULL
            );

        if (hFileMapping == NULL) writeErrFileRead(pfileInfo);
        pfileInfo->hFileMapping = hFileMapping;
        CloseHandle(hFile);
        pfileInfo = pfileInfo->pNext;
    }
    while (pfileInfo != pfileInfo1st);
}

static void
writeConsoleOutput(
HANDLE hConsoleOutput,
LPCWSTR lpCharacter,
DWORD nLength,
COORD dwWriteCoord)
{
    DWORD dwNumberOfCharsWritten;

    WriteConsoleOutputCharacterW(
        hConsoleOutput,
        lpCharacter,
        nLength,
        dwWriteCoord,
        &dwNumberOfCharsWritten
        );
}

static DWORD getAllocationGranularity()
{
    SYSTEM_INFO systemInfo;

    GetSystemInfo(&systemInfo);
    return systemInfo.dwAllocationGranularity;
}

static LONG WINAPI inPageErrorFilter(EXCEPTION_POINTERS *pExceptionInfo)
{
    if (
        pExceptionInfo->ExceptionRecord->ExceptionCode ==
        EXCEPTION_IN_PAGE_ERROR)
        writeErr(L"Error reading files.\n", NULL);
    return EXCEPTION_CONTINUE_SEARCH;
}

static const void *
getFileContents(const FILE_INFO *pfileInfo, ULONGLONG ullOffset, size_t size)
{
    const void *pvContents;
    {
        DWORD dwFileOffsetHigh  = ((DWORD *)&ullOffset)[1];
        DWORD dwFileOffsetLow   = ((DWORD *)&ullOffset)[0];

        pvContents =
            MapViewOfFile(
            pfileInfo->hFileMapping,
            FILE_MAP_READ,
            dwFileOffsetHigh,
            dwFileOffsetLow,
            size
            );
        if (pvContents == NULL) writeErrFileRead(pfileInfo);
    }
    return pvContents;
}

static BOOL
compareDwordPtr(const DWORD_PTR *padw1, const DWORD_PTR *padw2, size_t cElems)
{
    size_t iElem;

    for (iElem = 0; iElem < cElems; ++iElem)
        if (padw1[iElem] != padw2[iElem]) return FALSE;
    return TRUE;
}

static FILE_DIFF
compareContents(
HANDLE hConsoleOutput,
ULONGLONG ullSize,
COORD cursorPosition,
const FILE_INFO *pfileInfo1st)
{
    ULONGLONG cCyclesDone100;
    size_t uBufferSize;
    ULONGLONG cCycles;
    ULONGLONG cCyclesToDo;
    unsigned int uDone;
    size_t cDwordPtrs;
    ULONGLONG ullOffset;
    const FILE_INFO *pfileInfo;

    cCyclesDone100 = 0;
    uBufferSize = (size_t)getAllocationGranularity();
    cCycles = (ullSize - 1) / uBufferSize + 1;
    cCyclesToDo = cCycles;
    uDone = 0;
    cDwordPtrs = uBufferSize / sizeof(DWORD_PTR);
    ullOffset = 0;
    pfileInfo = pfileInfo1st;
    SetUnhandledExceptionFilter(inPageErrorFilter);
    for (;;)
    {
        const void *pvContents0;
        const FILE_INFO *pfileInfo0;

        if (hConsoleOutput != NULL)
        {
            unsigned int uNewDone = (unsigned int)(cCyclesDone100 / cCycles);

            if (uNewDone != uDone)
            {
                WCHAR szProgress[MAX_PROGRESS_DIGITS + 1];

                uDone = uNewDone;
                wsprintfW(
                    szProgress,
                    L"%" WSTR_MAX_PROGRESS_DIGITS L"u",
                    uDone
                    );
                writeConsoleOutput(
                    hConsoleOutput,
                    szProgress,
                    MAX_PROGRESS_DIGITS,
                    cursorPosition
                    );
            }
            cCyclesDone100 += 100;
        }
        if (cCyclesToDo == 1) // last cycle clause
        {
            uBufferSize = (size_t)(ullSize - ullOffset);
            cDwordPtrs =
                (uBufferSize + (sizeof(DWORD_PTR) - 1)) / sizeof(DWORD_PTR);
        }
        pvContents0 = getFileContents(pfileInfo, ullOffset, uBufferSize);
        pfileInfo0 = pfileInfo;
        for (;;)
        {
            FILE_INFO *pfileInfoNext;
            const void *pvContents;

            pfileInfoNext = pfileInfo->pNext;
            if (pfileInfoNext == pfileInfo0) break;

            pfileInfo = pfileInfoNext;
            pvContents = getFileContents(pfileInfo, ullOffset, uBufferSize);

            // This may read some bytes that would map to positions past the end
            // of the file: the bytes in excess are still in the same memory
            // page as the preceeding data and their contents are zeroed.
            if (
                !compareDwordPtr(
                (const DWORD_PTR *)pvContents0,
                (const DWORD_PTR *)pvContents,
                cDwordPtrs))
                return FILE_DIFF_CONTENTS;
            UnmapViewOfFile(pvContents);
        }
        --cCyclesToDo;
        if (cCyclesToDo == 0) return FILE_DIFF_NONE;

        ullOffset += uBufferSize;
        UnmapViewOfFile(pvContents0);
    }
}

#pragma warning(push)
#pragma warning(disable: 4701)

static FILE_DIFF
compareFiles(
FILE_INFO *pfileInfo1st,
BOOL bShowProgress,
BOOL bHeedReparsePoints)
{
    const FILE_INFO *pfileInfo;
    FILE_DIFF fileDiff;
    LONGLONG llSize0;

    // PASS 1: open files
    openFiles(pfileInfo1st, bHeedReparsePoints);

    // PASS 2: compare size of files
    llSize0 = getFileSize(pfileInfo1st);
    for (pfileInfo = pfileInfo1st; ; )
    {
        LONGLONG llSize;

        pfileInfo = pfileInfo->pNext;
        if (pfileInfo == pfileInfo1st) break;
    
        llSize = getFileSize(pfileInfo);
        if (llSize0 != llSize) return FILE_DIFF_SIZE;
    }

    // PASS 3: compare reparse points
    if (bHeedReparsePoints && !compareReparsePoints(pfileInfo1st))
        return FILE_DIFF_REPARSE_POINTS;

    if (llSize0 == 0) return FILE_DIFF_NONE;

    // PASS 4: create file mapping objects and close files
    createFileMappingsAndCloseFiles(pfileInfo1st);

    // PASS 5: compare contents of files
    {
        HANDLE hConsoleOutput;
        COORD cursorPosition;
        static const WCHAR sCompleted[] =
        {
            ZERO_PROGRESS_WCHAR_ARRAY, L'%', L' ', L'c', L'o', L'm', L'p', L'a',
            L'r', L'e', L'd', L'.'
        };

// This will prevent debug versions from raising a breakpoint interrupt
#ifdef _DEBUG
        memcpy(&cursorPosition, NULL, 0);
#endif

        if (!bShowProgress)
            hConsoleOutput = NULL;
        else
        {
            hConsoleOutput =
                CreateFileW(
                L"CONOUT$",
                GENERIC_READ | GENERIC_WRITE,
                FILE_SHARE_WRITE,
                NULL,
                OPEN_EXISTING,
                0,
                NULL
                );
            if (hConsoleOutput != NULL)
            {
                CONSOLE_SCREEN_BUFFER_INFO consoleScreenBufferInfo;

                if (
                    GetConsoleScreenBufferInfo(
                    hConsoleOutput,
                    &consoleScreenBufferInfo))
                {
                    cursorPosition = consoleScreenBufferInfo.dwCursorPosition;
                    writeConsoleOutput(
                        hConsoleOutput,
                        sCompleted,
                        _countof(sCompleted),
                        cursorPosition
                        );
                }
                else
                {
                    CloseHandle(hConsoleOutput);
                    hConsoleOutput = NULL;
                }
            }
        }
        fileDiff =
            compareContents(
            hConsoleOutput,
            llSize0,
            cursorPosition,
            pfileInfo1st
            );
        if (hConsoleOutput != NULL)
        {
            WCHAR sSpaces[_countof(sCompleted)];

            wmemset(sSpaces, L' ', _countof(sCompleted));
            writeConsoleOutput(
                hConsoleOutput,
                sSpaces,
                _countof(sCompleted),
                cursorPosition
                );
            CloseHandle(hConsoleOutput);
        }
    }
    return fileDiff;
}

#pragma warning(pop)

static BOOL
getCentisecs(long long llCountDiff, unsigned long long *pullCentisecs)
{
    const long long MAX_DIFF_COUNT =
        (0xffffffffffffffffu - 0xffffffffffffffffu / 200) / 100;
    LARGE_INTEGER freq;

    if (
        llCountDiff < 0 ||
        llCountDiff > MAX_DIFF_COUNT ||
        !QueryPerformanceFrequency(&freq))
        return FALSE;
    {
        unsigned long long ullFreq = freq.QuadPart;
        *pullCentisecs = (llCountDiff * 100 + ullFreq / 200) / ullFreq;
    }
    return TRUE;
}

#pragma warning(push)
#pragma warning(disable: 4701)

__declspec(noreturn) void __stdcall entryPoint()
{
    FILE_DIFF fileDiff;
    DWORD dwOptSet;
    BOOL bTimingAvailable;
    LARGE_INTEGER startCount, endCount;
    {
        FILE_INFO *pfileInfo1st;

        parseCommandLine(&pfileInfo1st, &dwOptSet);
        bTimingAvailable = (dwOptSet & OPT_TIMING) != 0;
        if (bTimingAvailable)
            bTimingAvailable = QueryPerformanceCounter(&startCount);
        fileDiff =
            compareFiles(
            pfileInfo1st,
            !(dwOptSet & OPT_QUIET),
            dwOptSet & OPT_HEED_REPARSE_POINTS
            );
        if (bTimingAvailable)
            bTimingAvailable = QueryPerformanceCounter(&endCount);
    }
    {
        UINT uExitCode;
        {
            PCWSTR szMessage;

            if (fileDiff == FILE_DIFF_NONE)
            {
                szMessage = L"The files have the same data.\n";
                uExitCode = EXIT_CODE_IDENTICAL;
            }
            else
            {
                switch (fileDiff)
                {
                case FILE_DIFF_SIZE:
                    szMessage = L"The files have different sizes.\n";
                    break;
                case FILE_DIFF_CONTENTS:
                    szMessage = L"The files have different contents.\n";
                    break;
                case FILE_DIFF_REPARSE_POINTS:
                    szMessage = L"The files differ in their reparse points.\n";
                default:
                    __assume(0);
                }
                uExitCode = EXIT_CODE_DIFFERENT;
            }
            if (!(dwOptSet & OPT_QUIET)) writeOut(szMessage, NULL);
        }
        if (dwOptSet & OPT_TIMING)
        {
            unsigned long long ullCentisecs;

            if (
                bTimingAvailable &&
                getCentisecs(
                endCount.QuadPart - startCount.QuadPart,
                &ullCentisecs))
            {
                PCWSTR pszTime;

                if (ullCentisecs != 0)
                {
                    // 18 characters for digits before decimal point
                    //  1 character for decimal point
                    //  2 characters for digits after decimal point
                    //  1 terminating null character
                    // ---------------------------------------------
                    // 22 characters required for buffer
                    WCHAR szTime[22];

                    pszTime =
                        formatUnitsAndCents(
                        ullCentisecs / 100,
                        ullCentisecs % 100,
                        szTime
                        );
                }
                else
                    pszTime = L"< 0.01";
                writeOut(L"File access time: ", pszTime, L" seconds.\n", NULL);
            }
            else
                writeOut(L"File access time not available.\n", NULL);
        }
        ExitProcess(uExitCode);
    }
}

#pragma warning(pop)

void __cdecl main()
{ }
